import util
import redivis


def test_list_variables():
    util.populate_test_data()
    table = util.get_table()
    variables = table.list_variables()
    print(variables)


def test_list_rows():
    util.populate_test_data()
    table = util.get_table()
    print(table.list_rows(max_results=10))


def test_to_dataframe():
    util.populate_test_data()
    table = util.get_table()
    print(table.to_dataframe(max_results=10))


def test_file_stream():
    from io import TextIOWrapper

    file = redivis.file("1934-e2htjwa76.AnL5YRjeUXtSlz6fhPVXog")
    lines_read = 0
    # stream = file.stream()

    # for chunk in file.stream():

    # print(stream.closed)
    with TextIOWrapper(file.stream()) as f:
        print(f.closed)
        while not f.closed:
            line = f.readline()
            if not line:
                break
            lines_read += 1
            print(lines_read)

    print(lines_read)


def test_to_arrow_batch_iterator():
    table = redivis.table("demo.ghcn_daily_weather_data.daily_observations")
    row_count = 0
    for batch in table.to_arrow_batch_iterator(1e6):
        row_count += batch.num_rows

    assert row_count == 1e6  # table.properties.get("numRows")


def test_selected_variables():
    util.populate_test_data()
    table = util.get_table()
    print(table.to_dataframe(max_results=10, variables=["ID"]))


def test_geo_dataframe():
    util.delete_test_dataset()
    util.populate_test_data("us_states.geojsonl")
    table = util.get_table()
    df = table.to_dataframe(max_results=100)
    print(df.dtypes)
    print(df)



import os
import concurrent.futures
import pathlib
from tqdm.auto import tqdm
from threading import Event


def download_files(
    self,
    path=None,
    *,
    overwrite=False,
    max_results=None,
    file_id_variable=None,
    progress=True,
    max_parallelization=os.cpu_count() * 5,
):
    files = self.list_files(max_results, file_id_variable=file_id_variable)

    if path is None:
        path = os.getcwd()
    else:
        path = os.path.expanduser(path)

    if progress:
        pbar_count = tqdm(total=len(files), leave=False, unit=" files")
        pbar_bytes = tqdm(
            unit="B",
            leave=False,
            unit_scale=True,
            bar_format="{n_fmt} ({rate_fmt})",
        )

    if not os.path.exists(path):
        pathlib.Path(path).mkdir(exist_ok=True, parents=True)

    def on_progress(bytes):
        nonlocal pbar_bytes
        pbar_bytes.update(bytes)

    def download(file, cancel_event):
        nonlocal progress
        nonlocal pbar_count
        # TODO: if overwriting file, first check if file is the same size, and if so check md5 hash, and skip if identical (need md5 in header)
        file.download(
            path,
            overwrite=overwrite,
            progress=False,
            on_progress=on_progress if progress else None,
            cancel_event=cancel_event,
        )
        if progress and not cancel_event.is_set():
            pbar_count.update()

    cancel_event = Event()
    # TODO: this should use async, not threads
    with concurrent.futures.ThreadPoolExecutor(
        max_workers=min(100, max_parallelization, len(files))
    ) as executor:
        futures = [executor.submit(download, file, cancel_event) for file in files]

        not_done = futures
        try:
            while not_done and not cancel_event.is_set():
                # next line 'sleeps' this main thread, letting the thread pool run
                freshly_done, not_done = concurrent.futures.wait(not_done, timeout=0.5)
                for future in freshly_done:
                    # Call result() on any finished threads to raise any exceptions encountered.
                    future.result()
        except KeyboardInterrupt:
            print("KeyboardInterrupt")
        finally:
            cancel_event.set()

            for future in not_done:
                # Only cancels futures that were never started
                future.cancel()

            # Shutdown all background threads, now that they should know to exit early.
            executor.shutdown(wait=True, cancel_futures=True)

    if progress:
        pbar_count.close()
        pbar_bytes.close()



from .Datasource import Datasource
from .Query import Query
from .Table import Table
from .Notebook import Notebook
from .Transform import Transform
from .Dataset import Dataset

from .Base import Base
from urllib.parse import quote as quote_uri
from ..common.api_request import make_request, make_paginated_request


class Workflow(Base):
    def __init__(self, name, *, user=None, properties=None):
        from .User import User  # avoid circular import

        if not user:
            if len(name.split(".")) == 2:
                user = User(name.split(".")[0])
                name = name.split(".")[-1]
            else:
                raise Exception(
                    "Invalid workflow specifier, must be the fully qualified reference if no user is specified"
                )

        if isinstance(user, str):
            user = User(user)

        self.user = user
        self.name = name

        self.qualified_reference = (
            properties["qualifiedReference"]
            if "qualifiedReference" in (properties or {})
            else (f"{self.user.name}.{self.name}")
        )
        self.scoped_reference = (
            properties["scopedReference"]
            if "scopedReference" in (properties or {})
            else f"{self.name}"
        )
        self.uri = f"/workflows/{quote_uri(self.qualified_reference, '')}"
        self.properties = properties

    def list_datasources(self, *, max_results=None):
        data_sources = make_paginated_request(
            path=f"{self.uri}/dataSources",
            page_size=100,
            max_results=max_results,
        )
        return [
            Datasource(data_source["id"], workflow=self, properties=data_source)
            for data_source in data_sources
        ]

    def list_tables(self, *, max_results=None):
        tables = make_paginated_request(
            path=f"{self.uri}/tables",
            page_size=100,
            max_results=max_results,
        )
        return [
            Table(table["name"], workflow=self, properties=table) for table in tables
        ]

    def list_notebooks(self, *, max_results=None):
        notebooks = make_paginated_request(
            path=f"{self.uri}/notebooks",
            page_size=100,
            max_results=max_results,
        )
        return [
            Notebook(name=notebook["name"], workflow=self, properties=notebook)
            for notebook in notebooks
        ]

    def list_transforms(self, *, max_results=None):
        transforms = make_paginated_request(
            path=f"{self.uri}/transforms",
            page_size=100,
            max_results=max_results,
        )
        return [
            Transform(name=transform["name"], workflow=self, properties=transform)
            for transform in transforms
        ]

    def exists(self):
        try:
            make_request(method="HEAD", path=self.uri)
            return True
        except Exception as err:
            if err.args[0]["status"] != 404:
                raise err
            return False

    def get(self):
        properties = make_request(method="GET", path=self.uri)
        update_properties(self, properties)
        return self

    def query(self, query):
        return Query(query, default_workflow=self.qualified_reference)

    def table(self, name):
        return Table(name, workflow=self)

    def notebook(self, name):
        return Notebook(name, workflow=self)

    def transform(self, name):
        return Transform(name, workflow=self)

    def datasource(self, source):
        return Datasource(source, workflow=self)


def update_properties(instance, properties):
    instance.properties = properties
    instance.qualified_reference = properties["qualifiedReference"]
    instance.scoped_reference = properties["scopedReference"]
    instance.name = properties["name"]
    instance.uri = properties["uri"]
